@@ -1,4 +1,4 @@
-/*====================================================================
+/* ====================================================================
  * Copyright (c) 2000 Gregory Trubetskoy.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -13,27 +13,20 @@
  *    the documentation and/or other materials provided with the
  *    distribution.
  *
- * 3. All advertising materials mentioning features or use of this
- *    software must display the following acknowledgment:
- *    "This product includes software developed by Gregory Trubetskoy
- *    for use in the mod_python module for Apache HTTP server 
- *    (http://www.modpython.org/)."
+ * 3. The end-user documentation included with the redistribution, if
+ *    any, must include the following acknowledgment: "This product 
+ *    includes software developed by Gregory Trubetskoy."
+ *    Alternately, this acknowledgment may appear in the software itself, 
+ *    if and wherever such third-party acknowledgments normally appear.
  *
  * 4. The names "mod_python", "modpython" or "Gregory Trubetskoy" must not 
  *    be used to endorse or promote products derived from this software 
  *    without prior written permission. For written permission, please 
- *    contact grisha@ispol.com.
+ *    contact grisha@modpython.org.
  *
  * 5. Products derived from this software may not be called "mod_python"
  *    or "modpython", nor may "mod_python" or "modpython" appear in their 
- *    names without prior written permission of Gregory Trubetskoy. For 
- *    written permission, please contact grisha@ispol.com..
- *
- * 6. Redistributions of any form whatsoever must retain the following
- *    acknowledgment:
- *    "This product includes software developed by Gregory Trubetskoy
- *    for use in the mod_python module for Apache HTTP server 
- *    (http://www.modpython.org/)."
+ *    names without prior written permission of Gregory Trubetskoy.
  *
  * THIS SOFTWARE IS PROVIDED BY GREGORY TRUBETSKOY ``AS IS'' AND ANY
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
@@ -51,7 +44,7 @@
  *
  * mod_python.c 
  *
- * $Id: mod_python.c,v 1.36.2.1 2000/12/12 14:56:51 gtrubetskoy Exp $
+ * $Id: mod_python.c,v 1.44 2000/12/18 19:50:03 gtrubetskoy Exp $
  *
  * See accompanying documentation and source code comments 
  * for details.
@@ -119,7 +112,7 @@ interpreterdata *get_interpreter_data(const char *name, server_rec *srv)
     interpreterdata *idata = NULL;
     
     if (! name)
-	name = GLOBAL_INTERPRETER;
+	name = MAIN_INTERPRETER;
 
     p = PyDict_GetItemString(interpreters, (char *)name);
     if (!p)
@@ -282,12 +275,6 @@ void python_init(server_rec *s, pool *p)
     if (! Py_IsInitialized()) 
     {
 
-	/* initialize types */
-/*	MpTable_Type.ob_type = &PyType_Type; */
-/* 	MpServer_Type.ob_type = &PyType_Type; */
-/* 	MpConn_Type.ob_type = &PyType_Type; */
-/* 	MpRequest_Type.ob_type = &PyType_Type; */
-
 	/* initialze the interpreter */
 	Py_Initialize();
 
@@ -501,6 +488,7 @@ static requestobject *get_request_object(request_rec *req)
     requestobject *request_obj;
     char *s;
     char s2[40];
+    PyThreadState *_save;
 
     /* see if there is a request object already */
     /* XXX there must be a cleaner way to do this, atol is slow? */
@@ -520,9 +508,9 @@ static requestobject *get_request_object(request_rec *req)
 	    /* take out the slash */
 	    req->path_info[i - 1] = 0;
 
-	    Py_BEGIN_ALLOW_THREADS;
+	    _save = PyEval_SaveThread();
 	    ap_add_cgi_vars(req);
-	    Py_END_ALLOW_THREADS;
+	    PyEval_RestoreThread(_save);
 	    request_obj = (requestobject *)MpRequest_FromRequest(req);
 
 	    /* put the slash back in */
@@ -531,9 +519,9 @@ static requestobject *get_request_object(request_rec *req)
 	} 
 	else 
 	{ 
-	    Py_BEGIN_ALLOW_THREADS;
+	    _save = PyEval_SaveThread();
 	    ap_add_cgi_vars(req);
-	    Py_END_ALLOW_THREADS;
+	    PyEval_RestoreThread(_save);
 	    request_obj = (requestobject *)MpRequest_FromRequest(req);
 	}
 	
@@ -567,9 +555,11 @@ static int python_handler(request_rec *req, char *handler)
     conf = (py_dir_config *) ap_get_module_config(req->per_dir_config, &python_module);
 
     /* is there a handler? */
-    if (! ap_table_get(conf->directives, handler)) {
-	if (! ap_table_get(req->notes, handler))
+    if (! ap_table_get(conf->directives, "PythonHandlerModule")) {
+	if (! ap_table_get(conf->directives, handler)) {
+	    if (! ap_table_get(req->notes, handler))
 	    return DECLINED;
+	}
     }
 
     /*
@@ -608,9 +598,15 @@ static int python_handler(request_rec *req, char *handler)
 	    
 	    s = ap_table_get(conf->dirs, handler);
 	    if (! s) {
-		/* this one must have been added via req.add_handler() */
-		char * ss = ap_pstrcat(req->pool, handler, "_dir", NULL);
-		s = ap_table_get(req->notes, ss);
+
+		/* are we using PythonHandlerModule? */
+		s = ap_table_get(conf->dirs, "PythonHandlerModule");
+		
+		if (! s) {
+		    /* this one must have been added via req.add_handler() */
+		    char * ss = ap_pstrcat(req->pool, handler, "_dir", NULL);
+		    s = ap_table_get(req->notes, ss);
+		}
 	    }
 	    if (strcmp(s, "") == 0)
 		interpreter = NULL;
@@ -644,7 +640,7 @@ static int python_handler(request_rec *req, char *handler)
     
     /* create thread state and acquire lock */
     tstate = PyThreadState_New(idata->istate);
-#ifdef WITH_THREAD  
+#ifdef WITH_THREAD 
     PyEval_AcquireThread(tstate);
 #else
     PyThreadState_Swap(tstate);
@@ -658,6 +654,7 @@ static int python_handler(request_rec *req, char *handler)
         {
 	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req,
 			  "python_handler: make_obcallback returned no obCallBack!");
+
 	    PyThreadState_Swap(NULL);
 	    PyThreadState_Delete(tstate);
 #ifdef WITH_THREAD
@@ -674,7 +671,7 @@ static int python_handler(request_rec *req, char *handler)
     if (interpreter)
 	ap_table_set(req->notes, "python_interpreter", interpreter);
     else
-	ap_table_set(req->notes, "python_interpreter", GLOBAL_INTERPRETER);
+	ap_table_set(req->notes, "python_interpreter", MAIN_INTERPRETER);
     
     /* create/acquire request object */
     request_obj = get_request_object(req);
@@ -687,15 +684,21 @@ static int python_handler(request_rec *req, char *handler)
 	request_obj->hstack = ap_pstrdup(req->pool, ap_table_get(conf->directives,
 								 handler));
     }
-    if ((s = ap_table_get(req->notes, handler))) {
-	if (request_obj->hstack) {
+    if ((s = ap_table_get(conf->directives, "PythonHandlerModule"))) {
+	if (request_obj->hstack)
 	    request_obj->hstack = ap_pstrcat(req->pool, request_obj->hstack,
 					     " ", s, NULL);
-	}
-	else {
+	else 
+	    request_obj->hstack = ap_pstrdup(req->pool, s);
+    }	
+    if ((s = ap_table_get(req->notes, handler))) {
+	if (request_obj->hstack)
+	    request_obj->hstack = ap_pstrcat(req->pool, request_obj->hstack,
+					     " ", s, NULL);
+	else
 	    request_obj->hstack = ap_pstrdup(req->pool, s);
-	}
     }
+
     /* 
      * Here is where we call into Python!
      * This is the C equivalent of
@@ -897,7 +900,6 @@ void python_cleanup_handler(void *data)
         {
 	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req,
 			  "python_cleanup_handler: make_obcallback returned no obCallBack!");
-
 	    PyThreadState_Swap(NULL);
 	    PyThreadState_Delete(tstate);
 #ifdef WITH_THREAD
@@ -914,7 +916,7 @@ void python_cleanup_handler(void *data)
     if (interpreter)
 	ap_table_set(req->notes, "python_interpreter", interpreter);
     else
-	ap_table_set(req->notes, "python_interpreter", GLOBAL_INTERPRETER);
+	ap_table_set(req->notes, "python_interpreter", MAIN_INTERPRETER);
     
     /* create/acquire request object */
     request_obj = get_request_object(req);
@@ -1243,6 +1245,10 @@ static const char *directive_PythonInitHandler(cmd_parms *cmd, void * mconfig,
 					       const char *val) {
     return python_directive(cmd, mconfig, "PythonInitHandler", val);
 }
+static const char *directive_PythonHandlerModule(cmd_parms *cmd, void * mconfig,
+					  const char *val) {
+    return python_directive(cmd, mconfig, "PythonHandlerModule", val);
+}
 static const char *directive_PythonPostReadRequestHandler(cmd_parms *cmd, 
 							  void * mconfig, 
 							  const char *val) {
@@ -1283,6 +1289,7 @@ void python_finalize(void *data)
 #ifdef WITH_THREAD
     PyEval_ReleaseLock();
 #endif
+
     /* create thread state and acquire lock */
     tstate = PyThreadState_New(idata->istate);
 #ifdef WITH_THREAD
@@ -1318,7 +1325,7 @@ static void PythonChildInitHandler(server_rec *s, pool *p)
      * Cleanups registered first will be called last. This will
      * end the Python enterpreter *after* all other cleanups.
      */
-    /* ap_register_cleanup(p, NULL, (void (*)(void *))Py_Finalize, ap_null_cleanup);*/
+
     ap_register_cleanup(p, NULL, python_finalize, ap_null_cleanup);
 
     /*
@@ -1594,6 +1601,14 @@ command_rec python_commands[] =
 	RAW_ARGS,
 	"Python logger handlers."
     },
+    {
+	"PythonHandlerModule",
+	directive_PythonHandlerModule,
+	NULL,
+	OR_ALL,
+	RAW_ARGS,
+	"A Python module containing handlers to be executed."
+    },
     {
 	"PythonNoReload",                 
 	directive_PythonNoReload,         
